/************************************************************************
 *   IRC - Internet Relay Chat, src/packet.c
 *   Copyright (C) 1990  Jarkko Oikarinen and
 *                       University of Oulu, Computing Center
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef lint
static  char sccsid[] = "@(#)packet.c	2.12 1/30/94 (C) 1988 University of Oulu, \
Computing Center and Jarkko Oikarinen";
static char *rcs_version = "$Id: packet.c,v 1.1.1.1 1997/09/29 15:35:27 db Exp $";

#endif
 
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "msg.h"
#include "h.h"
 
/*
** dopacket
**	cptr - pointer to client structure for which the buffer data
**	       applies.
**	buffer - pointr to the buffer containing the newly read data
**	length - number of valid bytes of data in the buffer
**
**      The buffer might be partially or totally zipped.
**      At the beginning of the compressed flow, it is possible that
**      an uncompressed ERROR message will be found.  This occurs when
**      the connection fails on the other server before switching
**      to compressed mode.
**
** Note:
**	It is implicitly assumed that dopacket is called only
**	with cptr of "local" variation, which contains all the
**	necessary fields (buffer etc..)
*/
int	dopacket(aClient *cptr, char *buffer, int length)
{
  Reg	char	*ch1;
  Reg	char	*ch2;
  Reg	char	*cptrbuf;
  aClient	*acpt = cptr->acpt;
#ifdef ZIP_LINKS
        int     zipped = 0;
#endif

  cptrbuf = cptr->buffer;
  me.receiveB += length; /* Update bytes received */
  cptr->receiveB += length;

  /* checking (cptr->receiveB & 0x400) is wrong if it went sometghing
  ** like from 0 to 0x803;  we could check (x & 0xfffff700) instead, but
  ** checking > or & should take the same number of cycles on reasonable
  ** processors anyway  -orabidoo
  */
  if (cptr->receiveB > 0x3ff)
    {
      cptr->receiveK += (cptr->receiveB >> 10);
      cptr->receiveB &= 0x03ff;	/* 2^10 = 1024, 3ff = 1023 */
    }

  if (acpt != &me)
    {
      acpt->receiveB += length;
      if (acpt->receiveB > 0x3ff)
	{
	  acpt->receiveK += (acpt->receiveB >> 10);
	  acpt->receiveB &= 0x03ff;
	}
    }
  else if (me.receiveB > 0x03ff)
    {
      me.receiveK += (me.receiveB >> 10);
      me.receiveB &= 0x03ff;
    }
  ch1 = cptrbuf + cptr->count;
  ch2 = buffer;

#ifdef	ZIP_LINKS
  if (cptr->flags2 & FLAGS2_ZIPFIRST)
    {
      if (*ch2 == '\n' || *ch2 == '\r')
	{
	  ch2++;
	  length--;
	}
	cptr->flags2 &= ~FLAGS2_ZIPFIRST;
    }

    if (length && (cptr->flags2 & FLAGS2_ZIP))
      {
	/* uncompressed buffer first */
	zipped = length;
	ch2 = unzip_packet(cptr, ch2, &zipped);
	length = zipped;
	zipped = 1;
	if (length == -1)
		return exit_client(cptr, cptr, &me,
				   "fatal error in unzip_packet(1)");
      }
#endif /* ZIP_LINKS */

  while (--length >= 0 && ch2)
    {
      register char g;
      g = (*ch1 = *ch2++);
      /*
       * Yuck.  Stuck.  To make sure we stay backward compatible,
       * we must assume that either CR or LF terminates the message
       * and not CR-LF.  By allowing CR or LF (alone) into the body
       * of messages, backward compatibility is lost and major
       * problems will arise. - Avalon
       */
      if ( g < '\16' && (g== '\n' || g == '\r'))
	{
	  if (ch1 == cptrbuf)
	    continue; /* Skip extra LF/CR's */
	  *ch1 = '\0';
	  me.receiveM += 1; /* Update messages received */
	  cptr->receiveM += 1;
	  if (cptr->acpt != &me)
	    cptr->acpt->receiveM += 1;
	  cptr->count = 0; /* ...just in case parse returns with
			   ** FLUSH_BUFFER without removing the
			   ** structure pointed by cptr... --msa
			   */
	  if (parse(cptr, cptrbuf, ch1) == FLUSH_BUFFER)
	    /*
	    ** FLUSH_BUFFER means actually that cptr
	    ** structure *does* not exist anymore!!! --msa
	    */
	    return FLUSH_BUFFER;
	  /*
	  ** Socket is dead so exit (which always returns with
	  ** FLUSH_BUFFER here).  - avalon
	  */
	  if (cptr->flags & FLAGS_DEADSOCKET)
	    return exit_client(cptr, cptr, cptr, (cptr->flags & FLAGS_SENDQEX) ?
			       "SendQ exceeded" : "Dead socket");
#ifdef	ZIP_LINKS
	  if ((cptr->flags2 & FLAGS2_ZIP) && (zipped == 0) &&
	      (length > 0))
	    {
	      /*
	      ** beginning of server connection, the buffer
	      ** contained PASS/CAPAB/SERVER and is now 
	      ** zipped!
	      ** Ignore the '\n' that should be here.
	      */
	      /* Checked RFC1950: \r or \n can't start a
	      ** zlib stream  -orabidoo
	      */

	      zipped = length;
	      if (zipped > 0 && (*ch2 == '\n' || *ch2 == '\r'))
		{
		  ch2++;
		  zipped--;
		}
	      cptr->flags2 &= ~FLAGS2_ZIPFIRST;
	      ch2 = unzip_packet(cptr, ch2, &zipped);
	      length = zipped;
	      zipped = 1;
	      if (length == -1)
		return exit_client(cptr, cptr, &me,
				   "fatal error in unzip_packet(2)");
	    }
#endif
	  ch1 = cptrbuf;
	}
      else if (ch1 < cptrbuf + (sizeof(cptr->buffer)-1))
	ch1++; /* There is always room for the null */
    }
  cptr->count = ch1 - cptrbuf;
  return 0;
}
